#include "raven_spn.h"
#include "spn_regs.h"
#include "init.h"

#include <cmath> // log

using spn =spn_regs<0x90000000>;

// huge arrays of XSPN input and referance data
extern std::array<uint8_t, 50000> input_data;
extern std::array<double, 10000> ref_data;
constexpr auto ln2 = std::log(2);

bool double_equals(double a, double b, double epsilon = 0.001)
{
    return std::abs(a - b) < epsilon;
}

void run_xspn(int in_addr, int out_addr) {
	spn::mode_reg() = 0;
	spn::input_length_reg() = 500; 		// each sample consists of 5 uint8 values
	spn::input_addr_reg() = in_addr;
	spn::output_addr_reg() = out_addr;
	spn::num_of_in_beats_reg() = 40;		// Number of AXI4 burst beats needed to load all input data
	spn::num_of_out_beats_reg() = 64;	// Number of AXI4 burst beats needed to store all result data
	spn::start_reg() = 1;
}

void load_input_data(int addr, const int k=0) {
    int * mem_base = (int *) addr;
	int offset = k * 5;

	for (size_t i = 0, j = 0; j < 625; i += 4, j++) { // load 5000 samples
		*(mem_base + j) = input_data.at(offset + i);
		*(mem_base + j) |= (input_data.at(offset + i + 1) << 8) & 0x0000FF00;
		*(mem_base + j) |= (input_data.at(offset + i + 2) << 16) & 0x00FF0000;
		*(mem_base + j) |= (input_data.at(offset + i + 3) << 24) & 0xFF000000;
	}
}

bool check_results(int addr, int k) {
	bool result = 0;
	int  step = 500;
	double *res_base = (double*) (addr);
	for (int i = 0; i < step; i++) {
		if (!double_equals(res_base[i] * ln2, ref_data.at(k + i))) {
			printf("XSPN ref %d comparison FAILED\n", k + i);
			result = 1;
		}
	}
    printf("Compared samples %d - %d with the reference\n", k, k+step);
	return result;
}

/*! \brief main function
 *
 */
int main() {
    platform_init();
    int ret_val = 0;

    spn::mode_reg() = 1;
    spn::start_reg() = 1;
	wait_for_interrupt();
    spn::interrupt_reg() = 1;
    printf("READOUT HW:0x%x\n", spn::readout_reg());

	int in_addr  = 0x800C0000;
	int out_addr = 0x800A0000;
    load_input_data(in_addr);

    int step = 500; // number of samples to be process at once
    for (int k = 0; k < 10000; k+=step) {

        run_xspn(in_addr, out_addr);
        printf("XSPN processes samples %d - %d\n", k, k+step);

        // pre-load data for the next run while HW accelerator is running
        if(k<(10000-step))
            load_input_data(in_addr, k+step);

        wait_for_interrupt();
    	spn::interrupt_reg() = 1;

		ret_val = check_results(out_addr, k);
	}

	return ret_val;
}
